// WekaClassify.java version 0.3
// Fomerly Known as SensevalClassify
// Java program to classify given Senseval test file
  
// Copyright (C) 2000-2003
// Ted Pedersen, University of Minnesota, Duluth
// tpederse@d.umn.edu
// Satanjeev Banerjee, University of Minnesota, Duluth
// bane0025@d.umn.edu
  
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
  
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
// ----------------------------------------------------------------------------
//				Start of program
// ----------------------------------------------------------------------------

import weka.core.*;
import weka.classifiers.*;
import weka.filters.*;
import java.io.*;
import java.util.*;
import java.util.zip.GZIPInputStream;
import java.lang.*;

public class WekaClassify 
{
    public static void main( String[] options ) throws Exception
    {	  
	WekaClassify sc = new WekaClassify();
	String classIndexString, objectInputFileName, testFileName;
	String learnerName;

	Classifier classifier;
	Instances test;
	int classIndex = -1;

	// See if user's not given any options
	if ( options.length == 0 ) 
	{
	    System.out.println(sc.minimalNotes());
	    return;
	}

	// See if help is wanted
	if ( Utils.getFlag('h', options) )
	{
	    System.out.println(sc.help());
	    return;
	}

	// See if version is wanted
	if ( Utils.getFlag('v', options))
	{
	    System.out.println(sc.version());
	    return;
	}

	// Get the model filename
	objectInputFileName = Utils.getOption('l', options);
	if (objectInputFileName.length() == 0)
	{
	    System.out.println("Model required to classify instances.");
	    return;
	}

	// Get the test filename
	testFileName = Utils.getOption('T', options);
	if (testFileName.length() == 0)
	{
	    System.out.println("No test file given.");
	    return;
	}
	
	// Get the class index, if provided
	classIndexString = Utils.getOption('c', options);
	if (classIndexString.length() != 0) 
	    classIndex = Integer.parseInt(classIndexString);

	Vector names = null;

	try {
	    names = sc.getNames( testFileName ); // function will open, read and close the file
	} catch (Exception e) {
	    throw e;
	}
	
	// Get the precision level
	String temp = Utils.getOption('p', options);
	int precision = 4;
	try { 
	    if (temp.length() > 0)
		precision = Integer.parseInt(temp.trim());
	} catch (NumberFormatException nfe) {
	    System.out.println("Could not recognize value of -p switch: \"" + temp + "\". Should be an integer value.");
	    return;
	}

	// Open the reader for 
	BufferedReader testReader = null;
	try {
	    testReader = new BufferedReader(new FileReader(testFileName));
	} catch (Exception e) {
	    throw new Exception("Can't open file " + testFileName);
	}

	// Load the test instances
	test = new Instances(testReader);
	if (classIndex != -1) {
	    test.setClassIndex(classIndex - 1);
	} else {
	    test.setClassIndex(test.numAttributes() - 1);
	}
	if (classIndex > test.numAttributes()) 
	{
	    System.out.println("Index of class attribute too large.");
	    return;
	}
	testReader.close();

	// Create the model input stream and create the classifier
	ObjectInputStream objectInputStream = null;
	try {
	    InputStream is = new FileInputStream(objectInputFileName);
	    if (objectInputFileName.endsWith(".gz")) {
		is = new GZIPInputStream(is);
	    }
	    objectInputStream = new ObjectInputStream(is);
	} catch (Exception e) {
	    throw new Exception("Can't open file " + objectInputFileName);
	}
	
	classifier = (Classifier) objectInputStream.readObject();
	objectInputStream.close();

	// now to get the classification of the test instances
	try {
	    System.out.println(sc.getDistributionFormatClassification(classifier, test, testReader, names, precision));
	} catch (Exception e) {
	    throw new Exception("Couldnt classify instances");
	}
    }

    private String getDistributionFormatClassification( Classifier classifier,
							Instances test, 
							BufferedReader testReader, 
							Vector names,
							int precision ) throws Exception
    {
	StringBuffer text = new StringBuffer();
	long precisionShifter = (long)Math.pow(10.0, (double)precision);

	for ( int iter = 0; iter < test.numInstances(); iter++ )
	{
	    Instance instance = test.instance(iter);
	    Instance withMissing = (Instance)instance.copy();
	    withMissing.setDataset(test);

	    if ( test.classAttribute().isNumeric() )
		return("Cannot create distribution for numeric output class");

	    text.append(names.elementAt(iter)+" ");
	    if ( classifier instanceof DistributionClassifier )
	    {
		double[] distribution = 
		    ((DistributionClassifier)classifier).distributionForInstance(withMissing);
		for ( int i = 0; i < test.classAttribute().numValues(); i++ )
		{
		    double roundedValue = (double)Math.round(distribution[i] * precisionShifter) / precisionShifter;
		    text.append(test.classAttribute().value(i) + "/" + roundedValue + " ");
		}
		if ( iter < test.numInstances() - 1 ) text.append("\n");
	    }
	    else 
	    {
		double predValue = ((Classifier)classifier).classifyInstance(withMissing);

		if (!Instance.isMissingValue(predValue)) 
		    text.append(test.classAttribute().value((int)predValue));
	    }
	}
	return text.toString();
    }

    private Vector getNames( String testFileName ) throws Exception
    {
	Instances tempInstances;
	Vector names = new Vector();
	BufferedReader testReader = null;
	try {
	    testReader = new BufferedReader( new FileReader(testFileName));
	} catch (Exception e) {
	    throw new Exception("Couldnt open test file in getNames()");
	}

	try {
	    tempInstances = new Instances(testReader, 1);
	} catch (Exception e) {
	    throw new Exception("Couldnt form temp instances in getNames()");
	}

	int index = 0;
	try {
	    String line = testReader.readLine();
	    
	    while(line != null )
	    {
		String temp;

		if ( line.indexOf('%') >= 0 ) 
		{
		    temp = line.substring(line.indexOf('%')+1);
		    if ( temp.length() == 0 ) 
			temp = String.valueOf(index);
		}
		else temp = String.valueOf(index);

		// trim leading blanks...
		while ( temp.indexOf(' ') == 0 )
		    temp = temp.substring(1);

		names.add((Object)temp);
		
		index++;
		line = testReader.readLine();
	    }
	} catch (Exception e) {
	    throw e;
	}

	try {
	    testReader.close();
	} catch (Exception e) {
	    throw new Exception("Couldnt close test file at getNames()");
	}
	
	return names;
    }

    private String help()
    {
	StringBuffer text = new StringBuffer();
	
	text.append("Usage: java WekaClassify -l <MODEL> -T <TESTFILE> [-phv]\n");
	text.append("Classifies data in given TESTFILE using classifer in MODEL, and outputs\n"); 
	text.append("probability distribution of possible senses in Senseval-2 output format.\n\n");
	text.append("Options:\n\n");
	text.append("    -l <MODEL>      : Use classifier in MODEL. MODEL should have been created\n");
	text.append("                      earlier using weka's -d option.\n\n");
	text.append("    -T <TESTFILE>   : Use instances in TESTFILE. Comments after instances are\n");
	text.append("                      used as instance names.\n\n");
	text.append("    -p              : Set the precision of output probabilities. Default = 4\n");
	text.append("                      places of decimal.\n\n");
	text.append("    -h              : Print this help message.\n\n");
	text.append("    -v              : Print out the version number.\n\n");
	return text.toString();
    }

    private String minimalNotes()
    {
	StringBuffer text = new StringBuffer();
	
	text.append("Usage: java WekaClassify -l <model> -T <testFile> [-phv]\n");
	text.append("Type java WekaClassify -h for further help.\n");
	
	return text.toString();
    }

    private String version()
    {
	StringBuffer text = new StringBuffer();
	
	text.append("WekaClassify - Version 0.3\n");
	text.append("A component of SenseTools 0.3\n");
	text.append("Copyright (C) 2001-2003: Ted Pedersen & Satanjeev Banerjee\n"); 
	text.append("Date of Last Update: May 01, 2003\n"); 

	return text.toString();
    }
}
