#!/usr/local/bin/perl -w

# read in a series of answer files that contain distributional info
# generated by different classifiers. Create an ensemble by summing
# the probabilities associated with each sense of each word and
# using the max probability to assign a sense

# Copyright (C) 2000-2001
# Ted Pedersen, University of Minnesota, Duluth
# tpederse@d.umn.edu
# Satanjeev Banerjee, University of Minnesota, Duluth
# bane0025@d.umn.edu
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
#-----------------------------------------------------------------------------
#                              Start of program
#-----------------------------------------------------------------------------

# some more information - 

# the following directory structure is assumed
# f0/word1.n/method1.answer
# f0/word1.n/method2.answer
# f1/word1.n/method1.answer
# f1/word1.n/method2.answer
#
# all answer files found in a subdirectory are considered
# 
# start this command by giving the feature directories that will
# be accessed - for example to create an ensemble of features
# directories f0 f1 f3, you would run :
#
# ensembleByDist.pl f0 f1 f3
#
# the program will enter the feature directory and then the word
# directory in order to find the answer files

# there will be an answer file created for each word that will
# contain the sum of the probabilities from each method

# we have to use commandline options, so use the necessary package!
use Getopt::Long;

# first check if no commandline options have been provided... in which case
# print out the usage notes!
if ( $#ARGV == -1 )
{
    &minimalUsageNotes();
    exit;
}

# now get the options!
GetOptions("verbose", "version", "help");

# if help has been requested, print out help!
if ( defined $opt_help )
{
    $opt_help = 1;
    &showHelp();
    exit;
}

# if version has been requested, show version!
if ( defined $opt_version )
{
    $opt_version = 1;
    &showVersion();
    exit;
}

foreach $featdir (@ARGV) 
{
    if ($opt_verbose) { print "Opened featdir $featdir\n"; }
    opendir(FEATDIR, $featdir) or die "can't open $featdir: $!";
    
    while (defined ($worddir = readdir(FEATDIR))) 
    {
	if ($worddir =~ /^\w/ ) 
	{ 
	    if ($opt_verbose) { print "  Opened worddir $worddir\n"; } 
	    opendir(WORDDIR, "$featdir/$worddir") or die "can't open $worddir: $!";
	    
	    # process the answer files associated with a word
	    while (defined ($file = readdir(WORDDIR))) 
	    {
		if ($file =~ /\.answer\.dist$/) 
		{
		    if ($opt_verbose) { print "    Opened file $file\n"; }
		    open (INFILE, "$featdir/$worddir/$file")  || 
			die "Can't open INFILE $!\n";
		    
		    while (<INFILE>) 
		    {
			($lexelt, $id, @answer) = split(/\s+/,$_);
			foreach $entry (@answer) 
			{
			    ($sense,$prob) = split(/\//,$entry);
			    $idx = $lexelt." ".$id." ".$sense;
			    $answer{$idx} += $prob; 
			}
		    }
		}                                        
	    }
	    closedir(WORDDIR);
	}
    }
    closedir(FEATDIR);
}

# for each id, display a list of sense/prob items
# id sense1/0.1 sense2/0.3 

$fileOpen = 0;

foreach $key (sort keys %answer) 
{
    ($lexelt, $id, $sense) = split(/ /,$key);
    $track{$lexelt}++;
    $track{$id}++;
    
    if ($track{$lexelt} == 1) 
    {
	if ($fileOpen) { print OUTFILE "\n"; $fileOpen = 0; }
	open (OUTFILE, ">$lexelt.ensemble.answer.dist")  || 
	    die "Can't open OUTFILE $!\n";
	$fileOpen = 1;
	$count = 0;
    }
    
    if (($track{$id} == 1) && ($count > 0)) 
    {
	print OUTFILE "\n";
    }
    
    if ($track{$id} == 1) 
    {
	print OUTFILE "$lexelt $id $sense\/$answer{$key}";
	$count++;
    }
    else 
    {
	print OUTFILE " $sense\/$answer{$key}";
    }
}

if ($fileOpen) { print OUTFILE "\n"; $fileOpen = 0; }

# function to output a minimal usage note when the user has not provided any
# commandline options
sub minimalUsageNotes
{
    print "Usage: ensembleByDist.pl [OPTIONS] DIR [[, DIR] ...]\n";
    askHelp();
}

# function to output help messages for this program
sub showHelp
{
    print "Usage: ensembleByDist.pl [OPTIONS] DIR [[, DIR] ...]\n\n";
    print "Reads in a series of answer files that contain distributional info generated\n";
    print "by different classifiers and creates an ensemble by summing the probabilities\n";
    print "associated with each sense of each word and using the max probability to\n";
    print "assign a sense. Answer files for words w1 w2 w3 should be in the\n";
    print "DIRectories in subdirectores thusly: DIR/w1/method1.answer, DIR/w2/method2.answer\n\n";
    print "OPTIONS:\n\n";
    print "  --version          Prints the version number.\n\n";
    print "  --help             Prints this help message.\n\n";
}

# function to output the version number
sub showVersion
{
    print "ensembleByDist.pl  -  Version 0.1\n";
    print "A component of the SenseTools Version 0.1\n";
    print "Copyright (C) 2000-2001, Ted Pedersen & Satanjeev Banerjee\n";
}

# function to output "ask for help" message when the user's goofed up!
sub askHelp
{
    print "Type ensembleByDist.pl --help for help.\n";
}


